﻿using Abide_AddOn_API;
using Abide_AddOn_API.Classes;
using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Windows.Forms;

namespace Abide.Classes
{
    public class AddOn_Manager : IAddOnHost
    {
        /// <summary>
        /// Gets and returns a list of addons in the current manager.
        /// </summary>
        public List<ITool> Tools
        {
            get;
            protected set;
        }
        /// <summary>
        /// Gets and returns a list of tab pages in the current manager.
        /// </summary>
        public List<ITabPage> TabPages
        {
            get;
            protected set;
        }
        /// <summary>
        /// Gets and returns a list of menu butons in the current manager.
        /// </summary>
        public List<IMenuButton> MenuButtons
        {
            get;
            protected set;
        }
        /// <summary>
        /// Gets and returns the count of the AddOns in the AddOns list
        /// </summary>
        public int AddOnCount
        {
            get { return Tools.Count; }
        }
        /// <summary>
        /// Gets and returns the count of the TabPages in the TabPage list
        /// </summary>
        public int TabPageCount
        {
            get { return TabPages.Count; }
        }
        /// <summary>
        /// Gets and returns the count of the MenuButtons in the MenuButton list
        /// </summary>
        public int MenuButtonCount
        {
            get { return MenuButtons.Count; }
        }
        /// <summary>
        /// Gets or sets the AddOn Manager's Halo Map Version support.
        /// </summary>
        public MapVersion VersionSupport
        {
            get;
            set;
        }

        public Abide_AddOn_API.Halo_2_Map.Halo2Map H2Map;
        public Abide_AddOn_API.Halo_2_Beta_Map.Halo2BetaMap H2BMap;

        public DebugXbox Xbox;
        public Abide_AddOn_API.IndexEntry SelectedEntry;
        public Abide_AddOn_API.Meta.HaloPlugin Plugin;

        public AddOn_Manager(Abide_AddOn_API.Halo_2_Map.Halo2Map Map, DebugXbox Xbox)
        {
            this.VersionSupport = MapVersion.Halo_2;
            this.H2Map = Map;
            this.Xbox = Xbox;
            this.Tools = new List<ITool>();
            this.TabPages = new List<ITabPage>();
            this.MenuButtons = new List<IMenuButton>();
        }
        public AddOn_Manager(Abide_AddOn_API.Halo_2_Beta_Map.Halo2BetaMap Map, DebugXbox Xbox)
        {
            this.VersionSupport = MapVersion.Halo_2b;
            this.H2BMap = Map;
            this.Xbox = Xbox;
            this.Tools = new List<ITool>();
            this.TabPages = new List<ITabPage>();
            this.MenuButtons = new List<IMenuButton>();
        }

        /// <summary>
        /// Gets an Addon from the assembly location.
        /// The assembly should be a DLL file (*.dll)
        /// </summary>
        /// <param name="AssemblyLocation">The location of the addon assembly.</param>
        /// <returns>If the return value is null, the function failed. If it is not null, the addon from the assembly is loaded.</returns>
        internal ITool[] LoadTools(string AssemblyLocation)
        {
            try
            {
                //Create Variables
                List<ITool> Tools = new List<ITool>();
                ITool Tool = null;
                Assembly toolAssembly = null;
                Type typeInterface = null;
                FileInfo Info = null;

                //Get FileInfo
                Info = new FileInfo(AssemblyLocation);

                //Check Extension
                if (Info.Extension == ".dll")
                {
                    //Get Assembly
                    toolAssembly = Assembly.LoadFile(AssemblyLocation);

                    foreach (Type toolType in toolAssembly.GetTypes())
                    {
                        if (toolType.IsPublic && !toolType.IsAbstract)
                        {
                            typeInterface = toolType.GetInterface(typeof(ITool).ToString(), true);

                            if (typeInterface != null)
                            {
                                try
                                {
                                    Tool = (ITool)Activator.CreateInstance(toolAssembly.GetType(toolType.ToString()));
                                    if (Tool != null && Tool.HaloMapVersion == VersionSupport)
                                    {
                                        Tool.Host = this;
                                        Tools.Add(Tool);
                                    }
                                }
                                catch (Exception ex)
                                {
                                    MessageBox.Show(string.Format(
                                        "Failed to load Tool: \r\n{0}\r\n------\r\nException Message: {1}",
                                      AssemblyLocation,
                                      ex.Message));
                                }
                            }
                        }
                    }
                }

                //Return
                return Tools.ToArray();
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format(
                    "Failed to load AddOn: \r\n{0}\r\n------\r\nException Message: {1}",
                  AssemblyLocation,
                  ex.Message));
            }

            return null;
        }
        /// <summary>
        /// Adds an addon to the list of addons.
        /// </summary>
        /// <param name="AssemblyLocation">The assembly's location.</param>
        /// <returns>Returns true if the method succeeds.</returns>
        public ITool[] AddTools(string AssemblyLocation)
        {
            //Create Variables
            ITool[] Tools = LoadTools(AssemblyLocation);

            //Check AddOn
            string AssemblyVersion = Assembly.GetExecutingAssembly().GetName().Version.ToString();
            foreach (ITool Tool in Tools)
            {
                if (Tool != null && Tool.AbideVersion == AssemblyVersion)
                {
                    this.Tools.Add(Tool);
                }
            }

            return Tools;
        }
        /// <summary>
        /// Gets an Addon from the assembly location.
        /// The assembly should be a DLL file (*.dll)
        /// </summary>
        /// <param name="AssemblyLocation">The location of the addon assembly.</param>
        /// <returns>If the return value is null, the function failed. If it is not null, the addon from the assembly is loaded.</returns>
        internal ITabPage[] LoadTabPages(string AssemblyLocation)
        {
            try
            {
                //Create Variables
                List<ITabPage> Pages = new List<ITabPage>();
                ITabPage TabPage = null;
                Assembly tabPageAssembly = null;
                Type typeInterface = null;
                FileInfo Info = null;

                //Get FileInfo
                Info = new FileInfo(AssemblyLocation);

                //Check Extension
                if (Info.Extension == ".dll")
                {
                    //Get Assembly
                    tabPageAssembly = Assembly.LoadFile(AssemblyLocation);

                    foreach (Type addOnType in tabPageAssembly.GetTypes())
                    {
                        if (addOnType.IsPublic && !addOnType.IsAbstract)
                        {
                            typeInterface = addOnType.GetInterface(typeof(ITabPage).ToString(), true);

                            if (typeInterface != null)
                            {
                                try
                                {
                                    TabPage = (ITabPage)Activator.CreateInstance(tabPageAssembly.GetType(addOnType.ToString()));
                                    if (TabPage != null && TabPage.HaloMapVersion == VersionSupport)
                                    {
                                        TabPage.Host = this;
                                        Pages.Add(TabPage);
                                    }
                                }
                                catch (Exception ex)
                                {
                                    MessageBox.Show(string.Format(
                                        "Failed to load TabPage: \r\n{0}\r\n------\r\nException Message: {1}",
                                      AssemblyLocation,
                                      ex.Message));
                                }
                            }
                        }
                    }
                }

                //Return
                return Pages.ToArray();
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format(
                    "Failed to load AddOn: \r\n{0}\r\n------\r\nException Message: {1}",
                  AssemblyLocation,
                  ex.Message));
            }

            return null;
        }
        /// <summary>
        /// Adds an addon to the list of addons.
        /// </summary>
        /// <param name="AssemblyLocation">The assembly's location.</param>
        /// <returns>Returns true if the method succeeds.</returns>
        public ITabPage[] AddTabPages(string AssemblyLocation)
        {
            //Create Variables
            ITabPage[] TabPages = LoadTabPages(AssemblyLocation);

            //Check AddOn
            string AssemblyVersion = Assembly.GetExecutingAssembly().GetName().Version.ToString();
            foreach (ITabPage TabPage in TabPages)
            {
                if (TabPage != null)
                {
                    this.TabPages.Add(TabPage);
                }
            }

            return TabPages;
        }
        /// <summary>
        /// Gets an Addon from the assembly location.
        /// The assembly should be a DLL file (*.dll)
        /// </summary>
        /// <param name="AssemblyLocation">The location of the addon assembly.</param>
        /// <returns>If the return value is null, the function failed. If it is not null, the addon from the assembly is loaded.</returns>
        internal IMenuButton[] LoadMenuButtons(string AssemblyLocation)
        {
            try
            {
                //Create Variables
                List<IMenuButton> MenuButtons = new List<IMenuButton>();
                IMenuButton MenuButton = null;
                Assembly toolAssembly = null;
                Type typeInterface = null;
                FileInfo Info = null;

                //Get FileInfo
                Info = new FileInfo(AssemblyLocation);

                //Check Extension
                if (Info.Extension == ".dll")
                {
                    //Get Assembly
                    toolAssembly = Assembly.LoadFile(AssemblyLocation);

                    foreach (Type menuButtonType in toolAssembly.GetTypes())
                    {
                        if (menuButtonType.IsPublic && !menuButtonType.IsAbstract)
                        {
                            typeInterface = menuButtonType.GetInterface(typeof(IMenuButton).ToString(), true);

                            if (typeInterface != null)
                            {
                                try
                                {
                                    MenuButton = (IMenuButton)Activator.CreateInstance(toolAssembly.GetType(menuButtonType.ToString()));
                                    if (MenuButton != null && MenuButton.HaloMapVersion == VersionSupport)
                                    {
                                        MenuButton.Host = this;
                                        MenuButtons.Add(MenuButton);
                                    }
                                }
                                catch (Exception ex)
                                {
                                    MessageBox.Show(string.Format(
                                        "Failed to load MenuButton: \r\n{0}\r\n------\r\nException Message: {1}",
                                      AssemblyLocation,
                                      ex.Message));
                                }
                            }
                        }
                    }
                }

                //Return
                return MenuButtons.ToArray();
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format(
                    "Failed to load AddOn: \r\n{0}\r\n------\r\nException Message: {1}",
                  AssemblyLocation,
                  ex.Message));
            }

            return null;
        }
        /// <summary>
        /// Adds an addon to the list of addons.
        /// </summary>
        /// <param name="AssemblyLocation">The assembly's location.</param>
        /// <returns>Returns true if the method succeeds.</returns>
        public IMenuButton[] AddMenuButtons(string AssemblyLocation)
        {
            //Create Variables
            IMenuButton[] MenuButtons = LoadMenuButtons(AssemblyLocation);

            //Check AddOn
            foreach (IMenuButton MenuButton in MenuButtons)
            {
                if (MenuButton != null)
                {
                    this.MenuButtons.Add(MenuButton);
                }
            }

            return MenuButtons;
        }
        /// <summary>
        /// Causes an addon to recieve an OnTagChanged event
        /// </summary>
        /// <param name="addOn">Addon to recieve the update.</param>
        public void UpdateTag(ITool addOn)
        {
            addOn.Api.ChangeTag(SelectedEntry);
        }
        /// <summary>
        /// Causes an addon to recieve the OnXboxChanged event
        /// </summary>
        /// <param name="addOn">Addon to recive the update</param>
        public void UpdateXbox(ITool addOn)
        {
            addOn.Api.ChangeXbox(Xbox);
        }
        /// <summary>
        /// Causes an tab page to recieve an OnTagChanged event
        /// </summary>
        /// <param name="tabPage">ITabPage to recieve the update.</param>
        public void UpdateTag(ITabPage tabPage)
        {
            tabPage.Api.ChangeTag(SelectedEntry);
        }
        /// <summary>
        /// Causes an tab page to recieve the OnXboxChanged event
        /// </summary>
        /// <param name="tabPage">ITabPage to recive the update</param>
        public void UpdateXbox(ITabPage tabPage)
        {
            tabPage.Api.ChangeXbox(Xbox);
        }

        /// <summary>
        /// Returns data requested by the AddOn.
        /// </summary>
        /// <param name="Command">The command sent by the child.</param>
        /// <param name="Sender">The child IAddOn or ITabPage.</param>
        /// <param name="Args">The passed arguments.</param>
        /// <returns></returns>
        object IAddOnHost.SendCommand(string Command, object Sender, params string[] Args)
        {
            IAddOn AddOnSender = Sender is IAddOn ? (IAddOn)Sender : null;

            if(AddOnSender != null)
                switch (AddOnSender.HaloMapVersion)
                {
                    case MapVersion.Halo_2:
                        //Check Halo Map...
                        if (H2Map == null)
                        {
                            throw new NullReferenceException("Halo 2 Map Instance is null!");
                        }
                        break;

                    case MapVersion.Halo_2b:
                        //Check Halo Map...
                        if (H2BMap == null)
                        {
                            throw new NullReferenceException("Halo 2 Beta Map Instance is null!");
                        }
                        break;

                    default:
                        throw new NotImplementedException("Map type not yet supported.");
                        break;
                }

            //Create Value
            object ReturnValue = null;

            //Handle Command
            switch (Command)
            {
                case "ReloadMap":
                    //Reload Map
                    if (AddOnSender != null)
                        switch (AddOnSender.HaloMapVersion)
                        {
                            case MapVersion.Halo_2:
                                H2Map.LoadMap(H2Map.IO.FileLocation); break;

                            case MapVersion.Halo_2b:
                                H2BMap.LoadMap(H2BMap.IO.FileLocation); break;
                        }
                    break;
                case "GetCurrentIndexEntry":
                    ReturnValue = SelectedEntry;
                    break;
                case "GetDebugXbox":
                    ReturnValue = Xbox;
                    break;
                case "GetMap":
                    if(AddOnSender != null)
                        switch (AddOnSender.HaloMapVersion)
                        {
                            case MapVersion.Halo_2:
                                ReturnValue = H2Map; break;

                            case MapVersion.Halo_2b:
                                ReturnValue = H2BMap; break;
                        }
                    break;
                case "GetCurrentPlugin":
                    ReturnValue = new Abide_AddOn_API.Meta.HaloPlugin(Plugin.FileName);
                    break;
                case "GetPlugin":
                    if (AddOnSender != null)
                        switch (AddOnSender.HaloMapVersion)
                        {
                            case MapVersion.Halo_2:
                                ReturnValue = new Abide_AddOn_API.Meta.HaloPlugin(Settings.H2_PluginsPath + "\\" + Args[0] + ".ent"); break;

                            case MapVersion.Halo_2b:
                                ReturnValue = new Abide_AddOn_API.Meta.HaloPlugin(Settings.H2B_PluginsPath + "\\" + Args[0] + ".ent"); break;
                        }
                    break;
            }

            //Return
            return ReturnValue;
        }

        /// <summary>
        /// Returns an IndexEntry from a native Halo 2 Index Entry and Halo 2 Map.
        /// </summary>
        /// <param name="Entry">The Native IndexEntry.</param>
        /// <param name="Map">The Native Halo_2_Map.</param>
        /// <returns>Abide_AddOn_API IndexEntry.</returns>
        Abide_AddOn_API.Halo_2_Map.IndexEntry GetEntry(Abide_AddOn_API.Halo_2_Map.IndexEntry Entry, Abide_AddOn_API.Halo_2_Map.Halo2Map Map)
        {
            //Continue?
            if (Entry == null)
                return null;

            //Create Entry
            Abide_AddOn_API.Halo_2_Map.OBJECT_ENTRY ObjEntry = new Abide_AddOn_API.Halo_2_Map.OBJECT_ENTRY();
            ObjEntry.id = Entry.Ident;
            ObjEntry.metaLength = Entry.Size;
            ObjEntry.offset = Entry.VirtualOffset;
            ObjEntry.tag = Entry.Class.ToCharArray();
            Array.Reverse(ObjEntry.tag);

            //Return
            Abide_AddOn_API.Halo_2_Map.IndexEntry ReturnValue = new Abide_AddOn_API.Halo_2_Map.IndexEntry(ObjEntry, Map.SecondaryMagic, Entry.FileName);
            return ReturnValue;
        }
    }
}
