﻿using Potential_Software.Win32;
using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Potential_Software.Potential_UI.Forms
{
    public class CustomWindow : Form
    {
        #region Properties
        /// <summary>
        /// The window's theme
        /// </summary>
        [Category("Appearance"), Description("The window theme.")]
        public CustomWindowTheme WindowTheme
        {
            get { return _theme; }
            set { _theme = value; }
        }
        [Category("Non Client Area"), Description("..."), Browsable(true), EditorBrowsable(EditorBrowsableState.Always)]
        public int NonClientTop
        {
            get { return _ncTop; }
            set { _ncTop = value; }
        }
        [Category("Non Client Area"), Description("..."), Browsable(true), EditorBrowsable(EditorBrowsableState.Always)]
        public int NonClientBottom
        {
            get { return _ncBottom; }
            set { _ncBottom = value; }
        }
        [Category("Non Client Area"), Description("..."), Browsable(true), EditorBrowsable(EditorBrowsableState.Always)]
        public int NonClientLeft
        {
            get { return _ncLeft; }
            set { _ncLeft = value; }
        }
        [Category("Non Client Area"), Description("..."), Browsable(true), EditorBrowsable(EditorBrowsableState.Always)]
        public int NonClientRight
        {
            get { return _ncRight; }
            set { _ncRight = value; }
        }
        #endregion
        #region Fields
        FormWindowState _prevState;
        CustomWindowTheme _theme;
        int _ncTop;
        int _ncBottom;
        int _ncLeft;
        int _ncRight;
        #endregion
        #region Events
        /// <summary>
        /// Occurs when the form's state changes
        /// </summary>
        [Category("Behaivor"), Description("Occurs when the form's state changes. (Maximized, Minimized, etc.)")]
        public event EventHandler FormStateChanged
        {
            add { onFormStateChanged += value; }
            remove { onFormStateChanged -= value; }
        }

        event EventHandler onFormStateChanged;
        #endregion
        #region Methods
        #region Constructor
        public CustomWindow()
        {
            _theme = new CustomWindowTheme();
            _ncTop = 8; _ncBottom = 8; _ncLeft = 8; _ncRight = 8;
            _prevState = base.WindowState;
            DoubleBuffered = true;
        }
        #endregion
        #region WndProc Message Hook
        protected override void WndProc(ref Message m)
        {
            //Design form normally...
            if (DesignMode)
            {
                base.WndProc(ref m);
                return;
            }

            //Preapre
            NCCALCSIZE_PARAMS nccsp = new NCCALCSIZE_PARAMS();
            RECT clnRect = new RECT();
            RECT wndRect = new RECT();
            Messages Msg = Messages.WM_NULL;
            Rectangle Bounds = new Rectangle();
            Rectangle ClientRect;
            IntPtr HDC;
            IntPtr CompatibleDC;
            IntPtr CompatibleBitmap;
            Point ScreenPt = Point.Empty;
            Point ClnPt = Point.Empty;

            //Set
            Msg = (Messages)m.Msg;
            User32.GetWindowRect(m.HWnd, out wndRect);
            Bounds = Rectangle.FromLTRB(wndRect.left, wndRect.top, wndRect.right, wndRect.bottom);
            ClientRect = new Rectangle(0, 0, Bounds.Width - (_ncLeft + _ncRight), Bounds.Height - (_ncTop + _ncBottom));

            //Check...
            switch (Msg)
            {
                #region WM_NCPAINT
                case Messages.WM_NCPAINT:
                    if (Bounds.Width == 0 || Bounds.Height == 0)
                        return;

                    HDC = User32.GetWindowDC(m.HWnd);
                    CompatibleDC = Gdi32.CreateCompatibleDC(HDC);
                    CompatibleBitmap = Gdi32.CreateCompatibleBitmap(HDC, Bounds.Width, Bounds.Height);

                    try
                    {
                        Gdi32.SelectObject(CompatibleDC, CompatibleBitmap);
                        Gdi32.BitBlt(CompatibleDC, 0, 0, Bounds.Width, Bounds.Height, HDC, 0, 0, WinUser.SRCCOPY);

                        using (Graphics g = Graphics.FromHdc(CompatibleDC))
                        {
                            g.ExcludeClip(new Rectangle(_ncLeft, _ncTop, ClientRect.Width, ClientRect.Height));

                            if (_theme != null)
                            {
                                _theme.OnPaintNonClientArea(new CustomWindowNonClientPaintEventArgs(new Rectangle(0, 0, Width, Height), g));
                            }
                        }

                        Gdi32.BitBlt(HDC, 0, 0, Bounds.Width, Bounds.Height, CompatibleDC, 0, 0, WinUser.SRCCOPY);
                    }
                    finally
                    {
                        Gdi32.DeleteObject(CompatibleBitmap);
                        Gdi32.DeleteDC(CompatibleDC);
                    }

                    User32.ReleaseDC(m.HWnd, HDC);

                    m.Result = IntPtr.Zero;
                    break;
                #endregion
                #region WM_NCCALCSIZE
                case Messages.WM_NCCALCSIZE:
                    if (m.WParam == WinUser.TRUE)
                    {
                        //Get our Non Client Calculate Size Parameters...
                        nccsp = (NCCALCSIZE_PARAMS)Marshal.PtrToStructure(m.LParam, typeof(NCCALCSIZE_PARAMS));

                        //Set Proposed Client Area
                        nccsp.rgrc0 = wndRect;
                        nccsp.rgrc0.top += _ncTop;
                        nccsp.rgrc0.bottom -= _ncBottom;
                        nccsp.rgrc0.left += _ncLeft;
                        nccsp.rgrc0.right -= _ncRight;

                        //Set back into memory
                        Marshal.StructureToPtr(nccsp, m.LParam, true);
                    }
                    else
                    {
                        //Get our Client Rectangle
                        clnRect = (RECT)Marshal.PtrToStructure(m.LParam, typeof(RECT));

                        //Set Proposed Client Area
                        clnRect = wndRect;
                        clnRect.top += _ncTop;
                        clnRect.bottom -= _ncBottom;
                        clnRect.left += _ncLeft;
                        clnRect.right -= _ncRight;

                        //Set back into memory
                        Marshal.StructureToPtr(clnRect, m.LParam, true);
                    }

                    m.Result = IntPtr.Zero;
                    break;
                #endregion
                #region WM_NCHITTEST
                case Messages.WM_NCHITTEST:
                    ScreenPt = new Point(m.LParam.ToInt32());
                    ClnPt = new Point(ScreenPt.X - Bounds.X, ScreenPt.Y - Bounds.Y);
                    m.Result = GetHitTestResult(ClnPt);
                    break;
                #endregion
                #region WM_NCACTIVATE
                case Messages.WM_NCACTIVATE:
                    //Check WParam...
                    if (m.WParam == (IntPtr)0)
                    {
                        m.Result = (IntPtr)1;
                    }

                    //Check WindowState
                    if (WindowState != FormWindowState.Minimized)
                    {
                        if (m.WParam != (IntPtr)0) //Activated
                        {
                            _theme.State = Forms.WindowState.Active;
                        }
                        else //Deactivated
                        {
                            _theme.State = Forms.WindowState.Inactive;
                        }

                        if (m.LParam != (IntPtr)(-1))
                        {
                            InvalidateAll();
                        }
                        else { DefWndProc(ref m); }
                    }
                    else { base.WndProc(ref m); }
                    break;
                #endregion

                default:
                    base.WndProc(ref m);
                    break;
            }
        }
        #endregion
        #region Overrides
        protected override void OnPaintBackground(PaintEventArgs e)
        {
            base.OnPaintBackground(e);

            e.Graphics.IntersectClip(new Rectangle(0, 0, ClientRectangle.Width, _theme.CaptionSize));

            if (_theme != null)
            {
                _theme.OnPaintTitleBar(e);
            }
        }
        protected override void OnSizeChanged(EventArgs e)
        {
            Size Prev = this.Size;
            base.OnSizeChanged(e);
            InvalidateAll();

            //Check for change in Window State...
            if (WindowState != _prevState)
            {
                OnFormStateChanged(e);
            }

            //Update State...
            _prevState = WindowState;

            Refresh();
        }
        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);
            InvalidateAll();
        }
        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);
            InvalidateAll();
        }
        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);
            Uxtheme.SetWindowTheme(this.Handle, string.Empty, string.Empty);
        }
        #endregion
        #region Virtual Methods
        /// <summary>
        /// Returns an IntPtr used to determine where the current hit test would return.
        /// </summary>
        /// <param name="ClientPoint">The Point inside the window (Client Point)</param>
        /// <returns>IntPtr containing the HitTest return value.</returns>
        protected virtual IntPtr GetHitTestResult(Point ClientPoint)
        {
            //Prepare Result
            IntPtr HRESULT = (IntPtr)WinUser.HTNOWHERE;

            //Check with NC Rectangles
            if (new Rectangle(0, 0, Bounds.Width, Bounds.Height).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTCLIENT;
            if (new Rectangle(0, 0, _ncLeft, _ncTop).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTTOPLEFT;
            if (new Rectangle(Width - _ncRight, 0, _ncRight, _ncTop).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTTOPRIGHT;
            if (new Rectangle(0, Bounds.Height - _ncBottom, _ncLeft, _ncBottom).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTBOTTOMLEFT;
            if (new Rectangle(Bounds.Width - _ncRight, Bounds.Height - _ncBottom, _ncRight, _ncBottom).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTBOTTOMRIGHT;
            if (new Rectangle(_ncLeft, 0, Bounds.Width - (_ncLeft + _ncRight), _ncTop).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTTOP;
            if (new Rectangle(0, _ncTop, _ncLeft, Bounds.Height - (_ncTop + _ncBottom)).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTLEFT;
            if (new Rectangle(Bounds.Width - _ncRight, _ncTop, _ncRight, Bounds.Height - (_ncTop + _ncBottom)).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTRIGHT;
            if (new Rectangle(_ncLeft, Bounds.Height - _ncBottom, Bounds.Width - (_ncLeft + _ncRight), _ncBottom).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTBOTTOM;
            if (new Rectangle(_ncLeft, _ncTop, Bounds.Width - (_ncRight + _ncBottom), _theme.CaptionSize).Contains(ClientPoint))
                HRESULT = (IntPtr)WinUser.HTCAPTION;

            return HRESULT; //Return...
        }
        /// <summary>
        /// Occurs when the form's state changes.
        /// </summary>
        /// <param name="e">The event args</param>
        protected virtual void OnFormStateChanged(EventArgs e)
        {
            if (onFormStateChanged != null)
                onFormStateChanged(this, e);
        }
        #endregion
        #region Functions
        /// <summary>
        /// Forces a repaint of the entire form. This includes the Non-Client area
        /// </summary>
        public void InvalidateAll()
        {
            if (!IsDisposed && IsHandleCreated)
            {
                //Redraw Frame...
                RECT rct = new RECT();
                User32.SetWindowPos(this.Handle, IntPtr.Zero, 0, 0, 0, 0,
                    WinUser.SWP_NOACTIVATE | WinUser.SWP_NOMOVE | WinUser.SWP_NOSIZE |
                    WinUser.SWP_NOZORDER | WinUser.SWP_NOOWNERZORDER | WinUser.SWP_FRAMECHANGED);
                User32.RedrawWindow(this.Handle, ref rct, IntPtr.Zero, WinUser.RDW_FRAME | WinUser.RDW_UPDATENOW | WinUser.RDW_INVALIDATE);

                //Invalidate...
                Invalidate();
            }
        }
        #endregion
        #endregion
    }
}
