﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace Potential_Software.Potential_UI.Forms
{
    public class CustomWindowTheme : Component
    {
        #region Events
        /// <summary>
        /// Occurs when the non client area has to be painted using this theme...
        /// </summary>
        [Category("Drawing"), Description("Occurs whenever the non client area needs to be painted using this theme.")]
        public event CustomWindowNonClientPaintEventHandler PaintNonClientArea
        {
            add { onPaintNonClientArea += value; }
            remove { onPaintNonClientArea -= value; }
        }
        /// <summary>
        /// Occurs when the title bar needs to be drawn
        /// </summary>
        [Category("Drawing"), Description("Occurs whenever the title bar needs to be drawn")]
        public event PaintEventHandler PaintTitleBar
        {
            add { onPaintTitleBar += value; }
            remove { onPaintTitleBar -= value; }
        }
        /// <summary>
        /// Occurs when the form's state changes from active to inactive or vise versa
        /// </summary>
        [Category("Drawing"), Description("Occurs when the form's state changes.")]
        public event EventHandler StateChanged
        {
            add { onStateChanged += value; }
            remove { onStateChanged -= value; }
        }

        event CustomWindowNonClientPaintEventHandler onPaintNonClientArea;
        event PaintEventHandler onPaintTitleBar;
        event EventHandler onStateChanged;
        #endregion
        #region Properties
        /// <summary>
        /// Controls the drawing of aero on the custom window
        /// </summary>
        [Category("Appearance"), Description("Controls whether the form will be drawn with Windows Aero theme."), DefaultValue(true)]
        public bool UseAero
        {
            get { return _useAero; }
            set { _useAero = value; }
        }
        /// <summary>
        /// Controls the height of the caption bar area
        /// </summary>
        [Category("Appearance"), Description("The height (in pixels) of the reserved caption bar size."), DefaultValue(22)]
        public int CaptionSize
        {
            get { return _capSize; }
            set { _capSize = value; }
        }
        /// <summary>
        /// The current state of the window
        /// </summary>
        [Browsable(false)]
        public WindowState State
        {
            get { return _state; }
            set { _state = value; OnStateChanged(new EventArgs()); }
        }
        /// <summary>
        /// Gets or sets the width of the resizing border
        /// </summary>
        [Browsable(false)]
        public int BorderSize
        {
            get { return _borderSize; }
            set { _borderSize = value; }
        }
        #endregion
        #region Fields
        WindowState _state = WindowState.Active;
        CustomWindow _parent = null;
        bool _useAero = true;
        int _capSize = 22;
        int _borderSize = 8;
        #endregion
        #region Methods
        #region Constructor
        public CustomWindowTheme() { }
        public CustomWindowTheme(CustomWindow ParentWindow) { this._parent = ParentWindow; }
        #endregion
        #region Virtual Methods
        /// <summary>
        /// Occurs when the non client area needs to be painted using this theme.
        /// </summary>
        /// <param name="e">The NonClientPaint event arguments</param>
        public virtual void OnPaintNonClientArea(CustomWindowNonClientPaintEventArgs e)
        {
            //Prepare
            Rectangle Caption = new Rectangle();
            Brush BackColorBrush = null;
            Brush CaptionBrush = null;
            Pen BorderPen = null;

            //Check State
            if (State == WindowState.Active)
            {
                Caption = new Rectangle(0, 0, e.Bounds.Width, _capSize);
                BackColorBrush = new SolidBrush(SystemColors.ActiveCaption);
                CaptionBrush = new LinearGradientBrush(Caption, SystemColors.GradientActiveCaption, 
                    SystemColors.ActiveCaption, LinearGradientMode.Vertical);
                BorderPen = new Pen(SystemColors.ActiveBorder);
            }
            else
            {
                Caption = new Rectangle(0, 0, e.Bounds.Width, _capSize );
                BackColorBrush = new SolidBrush(SystemColors.InactiveCaption);
                CaptionBrush = new LinearGradientBrush(Caption, SystemColors.GradientInactiveCaption, 
                    SystemColors.InactiveCaption, LinearGradientMode.Vertical);
                BorderPen = new Pen(SystemColors.InactiveBorder);
            }

            //Draw...
            e.Graphics.FillRectangle(BackColorBrush, e.Bounds);
            e.Graphics.FillRectangle(CaptionBrush, Caption);
            e.Graphics.DrawRectangle(BorderPen, e.Bounds);

            //Cleanup
            BackColorBrush.Dispose();
            CaptionBrush.Dispose();
            BorderPen.Dispose();

            //Events...
            if (onPaintNonClientArea != null)
                onPaintNonClientArea(this, e);
        }
        /// <summary>
        /// Occurs when the title bar needs to be drawn.
        /// </summary>
        /// <param name="e">The Paint event arguments</param>
        public virtual void OnPaintTitleBar(PaintEventArgs e)
        {
            //Events...
            if (onPaintTitleBar != null)
            {
                onPaintTitleBar(this, e);
            }
            else
            {
                //Prepare
                Rectangle Caption = new Rectangle(-_borderSize, -_borderSize, e.ClipRectangle.Width + (_borderSize * 2), _capSize + _borderSize);
                Brush CaptionBrush = null;

                //Check State
                if (State == WindowState.Active)
                {
                    CaptionBrush = new LinearGradientBrush(Caption, SystemColors.GradientActiveCaption,
                        SystemColors.ActiveCaption, LinearGradientMode.Vertical);
                }
                else
                {
                    CaptionBrush = new LinearGradientBrush(Caption, SystemColors.GradientInactiveCaption,
                        SystemColors.InactiveCaption, LinearGradientMode.Vertical);
                }

                //Draw...
                e.Graphics.FillRectangle(CaptionBrush, Caption);

                //Cleanup
                CaptionBrush.Dispose();
            }
        }
        /// <summary>
        /// Occurs when the window's state changes
        /// </summary>
        /// <param name="e">The event arguments</param>
        public virtual void OnStateChanged(EventArgs e)
        {
            if (onStateChanged != null)
                onStateChanged(this, e);
        }
        #endregion
        #region Delegates
        public delegate void CustomWindowNonClientPaintEventHandler(object sender, CustomWindowNonClientPaintEventArgs e);
        #endregion
        #endregion
    }

    public class CustomWindowNonClientPaintEventArgs : EventArgs, IDisposable
    {
        public Rectangle Bounds { get { return _bounds; } }
        public Graphics Graphics { get { return _gfx; } }

        Rectangle _bounds;
        Graphics _gfx;

        public CustomWindowNonClientPaintEventArgs(Rectangle Bounds, Graphics Gfx)
        {
            _bounds = Bounds;
            _gfx = Gfx;
        }
        public void Dispose()
        {
            _gfx.Dispose();
        }
    }
    public enum WindowState
    {
        Active,
        Inactive
    };
}
